import glpk, random, sys

# Conjunction of disjunction instances can defined as a list of tuples
# of integers.  Each integer indexes into some variable (counting from
# 1).  A negative index indicates the "not" of the appropriate
# variables.  The formula is the result of "or"ing all the quantities
# within the tuples, and the "and"ing across all tuples.  For one
# example, a 3-SAT formula would be a case where all the tuples were
# of length 3.

# For example, the expression
#   ( !x2 \/ !x3 \/ x4 ) /\ ( !x1 \/ x2 \/ x4 ) /\ ( x2 \/ !x3 \/ !x4 )
# would have representation
#   [(-2, -3, 4), (1, 2, 4), (2, -3, -4)]

# An interesting problem is the SAT problem, where we try to find an
# assignment of truth values to each variable so that a given
# expression becomes true.  This module has a SAT solver.

def generate_cnf(numvars, numclauses, perclause=3, rgen=random):
    """Generate a random CNF expression.

    Generates a random CNF formula over the given number of variables,
    clauses, and n, using the given random generator.  The argument
    numvars serves as the number of variables.  Numclauses serves as
    the number of clauses that are generated.  Perclause serves as the
    number of literals per clause."""
    if (numvars < 3): raise ValueError("need more than 3 variables")
    return [tuple(rgen.choice((-v,v)) for v in rgen.sample(
        range(1,numvars+1),3)) for c in range(numclauses)]

def solve_sat(expression):
    """Attempts to satisfy a formula of conjunction of disjunctions.

    If there are n variables in the expression, this will return a
    list of length n, all elements booleans.  The truth of element i-1
    corresponds to the truth of variable i.

    If no satisfying assignment could be found, None is returned."""
    # Trivial boundary case.
    if len(expression)==0: return []
    # Now many variables do we have?
    numvars = max(max(abs(v) for v in clause) for clause in expression)

    # Construct an empty linear program.
    lp = glpk.LPX()

    # The output GLPK produces is rather annoying.
    glpk.env.term_on = False

    # We want twice as many columns (LP variables) as there are
    # logical variables in the expression: one column for each
    # positive literal, and one column for each negative literal.  A
    # literal is "true" if its column holds 1, false if 0.
    lp.cols.add(2*numvars)

    # Bound all columns (LP variables) to have value between 0 and 1.
    for col in lp.cols:
        col.bounds = 0.0, 1.0

    # Let us suppose that literals x_i and !x_i correspond to columns
    # 2*i-2 and 2*i-1 respectively.  (So, columns 0 and 1 correspond
    # to the truth of x_1 and !x_1, respectively, columns 2 and 3 to
    # the truth of x_2 and !x_2, etc.)

    # Let's just define a helper function that will perform the
    # mapping of literal identifier to column number.  (So 1 maps to
    # 0, -1 to 1, 2 to 2, -2 to 3, 3 to 4, etc.)
    def lit2col(lit):
        return 2*lit-2 if lit>0 else 2*(-lit)-1

    # Here we do two things: we set the names of the column variables
    # appropriately (not required, but fun), and define a row in a
    # constraint matrix indicating that a literal and its complement
    # must sum to exactly 1 since exactly one of the two literals must
    # be true.
    for i in range(1, numvars+1):
        lp.cols[lit2col( i)].name =  'x_%d'%i
        lp.cols[lit2col(-i)].name = '!x_%d'%i
        lp.rows.add(1)
        lp.rows[-1].matrix = [(lit2col(i), 1.0), (lit2col(-i), 1.0)]
        lp.rows[-1].bounds = 1.0

    # We want to find an assignment of such variables so that each
    # disjunction (or-ing) is true.  A disjunction is true if one of
    # its literals is true.  Literal truth corresponds to 1 in the
    # corresponding column, so we can represent truth of a disjunction
    # by defining a constraint saying the sum of its literals must be
    # at least 1.  We do this for all clauses.
    for clause in expression:
        lp.rows.add(1)
        lp.rows[-1].matrix = [(lit2col(lit), 1.0) for lit in clause]
        lp.rows[-1].bounds = 1, None

    # Now we have the LP built.  Run the simplex algorithm.
    retval = lp.simplex()

    # If our iteration terminated prematurely, or if we do not have an
    # optimal solution, assume we have failed.
    if retval != None: return None
    if lp.status != 'opt': return None

    # Now switch this from a linear continuous problem to a
    # mixed-integer problem, and say we want all column variables
    # integer as well (that is, not just between 0 and 1, but exactly
    # 0 and 1).
    for col in lp.cols:
        col.kind = int

    # Attempt to solve this MIP problem with the MIP solver.
    retval = lp.integer()

    # Again, only returns non-None on failure.
    if retval != None: return None
    if lp.status != 'opt': return None

    # We want to return a list of boolean quantities, where the first
    # element is True iff x_1 is true, the second element is True iff
    # x_2 is true, and so on.  Variable truth corresponds to the
    # literals represented in the even columns.  So, we can just go
    # over all of the even columns (every 2 counting from 0), see if
    # the value is close to 1, and use that as our variable
    # assignment.
    return [col.value > 0.99 for col in lp.cols[::2]]

def verify(expression, assignment):
    """Return the truth of an expression given a variable truth assignment."""
    # Each clause must be true.
    for clause in expression:
        # For a disjunctive clause to be true, at least one of its
        # literals must be true.
        lits_true = 0
        for lit in clause:
            if (lit>0) == (assignment[abs(lit)-1]):
                lits_true += 1
        if lits_true == 0:
            return False
    return True

if __name__=='__main__':
    # If we have arguments, use them to seed the random generator.
    if len(sys.argv) > 1:
        random.seed(tuple(sys.argv))
    expression = generate_cnf(4, 16)
    assignment = solve_sat(expression)

    if assignment:
        print('Assignment found.  It is')
        print('valid.' if verify(expression, assignment) else 'invalid.')
    else:
        print('No assignment found.')



from setuptools import setup, Extension
from setuptools_scm import get_version
import os
import sys

useparams = False

sources = 'glpk 2to3 lp barcol bar obj util kkt tree environment'
source_roots = sources.split()
if useparams:
    source_roots.append('params')

# This build process will not work with anything prior to GLPK 4.16,
# since there were many notable changes in GLPK including,
# importantly, something which actually contains the version number.

pkgdirs = [] # incdirs and libdirs get these
libs = ["glpk"]
defs = []
incdirs = []
libdirs = []

def append_env(L, e):
    v = os.environ.get(e)
    if v and os.path.exists(v):
        L.append(v)

append_env(pkgdirs, "GLPK")

# Hack up sys.argv
unprocessed = []
for arg in sys.argv[1:]:
    if arg.startswith("--with-glpk="):
        pkgdirs.append(arg.split("=", 1)[1])
        continue
    unprocessed.append(arg)
sys.argv[1:] = unprocessed

for pkgdir in pkgdirs:
    incdirs.append(os.path.join(pkgdir, "include"))
    libdirs.append(os.path.join(pkgdir, "lib"))

is_windows = sys.platform.startswith("win")
version_fmt_str = '"{}"' if not is_windows else '"\\"{}\\""'
defs.append(('VERSION_NUMBER', version_fmt_str.format(get_version())))

if useparams:
    defs.append(('USEPARAMS', None))

runtime_library_dirs = libdirs if not is_windows else []

# Now, finally, define that module!
pyglpk_ext = Extension(
    'glpk', [os.path.join('src', r+'.c') for r in source_roots],
    libraries=libs, include_dirs=incdirs,
    library_dirs=libdirs, define_macros=defs,
    runtime_library_dirs=runtime_library_dirs)

setup(
    use_scm_version=True,
    ext_modules=[pyglpk_ext]
)



"""Tests for the environment."""

# These tests are somewhat different from the other test suites
# insofar as they directly muck with global settings of the underlying
# GLPK library, and thus have some chance of affecting other tests,
# and being affected by the workings of other tests.

import glpk
import random
import gc
import unittest


class MemoryTestCase(unittest.TestCase):
    """Tests for the environment memory variables."""
    def setUp(self):
        pass

    def doNottestEmpty(self):
        """Test that there are currently 0 bytes allocated."""
        # This test doesn't really work for some reason.
        gc.collect()  # Ensure any leftover LPX objects are cleaned up.
        self.assertEqual(glpk.env.blocks, 0)
        self.assertEqual(glpk.env.bytes, 0)

    def testPeakIsBound(self):
        """Tests that the peak is always greater than current."""
        self.assertTrue(glpk.env.blocks <= glpk.env.blocks_peak)
        self.assertTrue(glpk.env.bytes <= glpk.env.bytes_peak)

    def testIncrease(self):
        """Test that creating a new LPX increases the memory allocated."""
        gc.collect()
        bytes_first = glpk.env.bytes
        lp = glpk.LPX()
        lp.rows.add(2)
        lp.cols.add(2)
        lp.name = 'foo'
        self.assertTrue(glpk.env.bytes > bytes_first)
        del lp
        gc.collect()
        self.assertEqual(glpk.env.bytes, bytes_first)

    @unittest.skipIf(glpk.env.version < (4, 19), 'TODO')
    def testMemLimitNegative(self):
        """Test that negative memory limits are not permitted."""
        with self.assertRaises(ValueError):
            glpk.env.mem_limit = -1
        with self.assertRaises(ValueError):
            glpk.env.mem_limit = -500

    @unittest.skipIf(glpk.env.version < (4, 19), 'TODO')
    def testMemLimit(self):
        """Test setting the memory limit."""
        limits = [5, 0, 10, None, 2000]
        self.assertEqual(glpk.env.mem_limit, None)
        for limit in limits:
            glpk.env.mem_limit = limit
            self.assertEqual(glpk.env.mem_limit, limit)
        del glpk.env.mem_limit
        self.assertEqual(glpk.env.mem_limit, None)

    @unittest.skipIf(glpk.env.version < (4, 19), 'TODO')
    def testMemLimitType(self):
        """Test that only integer memory limits are permitted."""
        limits = ['hi!', 3.14159, complex(2, 3), {'foo': 'bar'}, (23,)]
        for limit in limits:
            with self.assertRaises(TypeError):
                glpk.env.mem_limit = limit


class TerminalTest(unittest.TestCase):
    """Tests for the terminal output settings."""
    def setUp(self):
        # Store the term_on value.
        self.term_on = glpk.env.term_on
        # One area where GLPK produces output is in its settings.
        lp = self.lp = glpk.LPX()
        lp.rows.add(1)
        lp.cols.add(2)
        lp.cols[0].name = 'x'
        lp.cols[1].name = 'y'
        for c in lp.cols:
            c.bounds = 0, 1
        lp.obj[:] = [1, 1]
        lp.obj.maximize = True
        lp.rows[0].matrix = [0.5, 1.0]
        lp.rows[0].bounds = None, 1

    def tearDown(self):
        # Restore term_on to whatever value it had when we started.
        glpk.env.term_on = self.term_on
        del glpk.env.term_hook

    def testTerminalRedirect(self):
        """Test setting the hook function simply."""
        glpk.env.term_on = True
        self.char_count = 0

        def count_hook(s):
            self.char_count += len(s)

        glpk.env.term_hook = count_hook
        self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
        self.assertTrue(self.char_count > 0)

    def testTerminalOnOff(self):
        """Test setting the terminal on and off."""
        glpk.env.term_on = True
        self.char_count = 0

        def count_hook(s):
            self.char_count += len(s)

        glpk.env.term_hook = count_hook
        self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
        self.assertTrue(self.char_count > 0)
        # Now turn the terminal output off.
        glpk.env.term_on = False
        oldcount = self.char_count
        self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
        if glpk.env.version < (4, 21) or glpk.env.version > (4, 26):
            # GLPK versions 4.21 through 4.26 did not fully respect
            # the environment settings.
            self.assertEqual(oldcount, self.char_count)
        # Now turn the terminal output back on.
        glpk.env.term_on = True
        self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
        self.assertTrue(self.char_count > oldcount)

    def testMultipleTerminalDirect(self):
        """Test setting the hook function multiple times."""
        glpk.env.term_on = True
        numhooks = 3
        self.char_counts = [0]*numhooks

        def makehook(which):
            def thehook(s):
                self.char_counts[which] += len(s)
            return thehook

        self.hooks = [makehook(i) for i in range(numhooks)]
        # Randomly reset the hooks.
        rgen = random.Random(10)
        for trial in range(100):
            whichhook = rgen.randint(0, numhooks-1)
            glpk.env.term_hook = self.hooks[whichhook]
            old_counts = self.char_counts[:]
            self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
            # Test that only the count associated with this hook has changed.
            for w in range(numhooks):
                if w == whichhook:
                    self.assertTrue(self.char_counts[w] > old_counts[w])
                else:
                    self.assertEqual(self.char_counts[w], old_counts[w])

    def testErrorInTermHook(self):
        """Test that errors in the terminal hook are ignored."""
        glpk.env.term_on = True

        def error_hook(s):
            1.0 / 0.0
        glpk.env.term_hook = error_hook
        self.lp.simplex(msg_lev=glpk.LPX.MSG_ALL)
        self.assertEqual(self.lp.status, 'opt')

    def testGetTermOn(self):
        """Test getting/setting term_on."""
        glpk.env.term_on = True
        self.assertTrue(glpk.env.term_on)
        with self.assertRaises(TypeError) as cm:
            glpk.env.term_on = 'foo'
        self.assertIn('term_on must be set with bool', str(cm.exception))

    def testGetTermHook(self):
        """Test getting the terminal hook."""
        self.assertTrue(glpk.env.term_hook is None)

        def noop_hook(s):
            pass
        glpk.env.term_hook = noop_hook
        self.assertEqual(noop_hook, glpk.env.term_hook)
